@@ -1,28 +1,15 @@
-use crate::NPage;
-use crate::BUFFER;
-use compact_str::CompactString;
+use crate::{new_page, pydict_to_json_value, NPage, BUFFER};
 use indexmap::IndexMap;
 use pyo3::prelude::*;
+use pyo3::types::PyDict;
+use spider::compact_str::CompactString;
+use spider::configuration::WaitForIdleNetwork;
+use spider::tokio::select;
 use spider::tokio::task::JoinHandle;
 use spider::utils::shutdown;
 use std::time::Duration;
 
-#[pyfunction]
-fn crawl(py: Python, url: String, raw_content: Option<bool>) -> PyResult<&PyAny> {
-  pyo3_asyncio::tokio::future_into_py(py, async move {
-    let w = crate::crawl(url, raw_content).await;
-
-    Ok(w)
-  })
-}
-
-#[pymodule]
-fn spider_rs(_py: Python<'_>, m: &PyModule) -> PyResult<()> {
-  m.add_function(wrap_pyfunction!(crawl, m)?)?;
-  Ok(())
-}
-
-/// a website holding the inner spider::website::Website from Rust fit for python.
+/// A website holding the inner spider::website::Website from Rust fit for python.
 #[pyclass]
 pub struct Website {
   /// the website from spider.
@@ -33,27 +20,20 @@ pub struct Website {
   crawl_handles: IndexMap<u32, JoinHandle<()>>,
   /// do not convert content to UT8.
   raw_content: bool,
-  /// the data collected.
-  collected_data: Box<Vec<u8>>,
   /// is the crawl running in the background.
   running_in_background: bool, // /// the file handle for storing data
-                               // file_handle: Option<spider::tokio::fs::File>,
-}
-
-struct PageEvent {
-  pub page: NPage,
 }
 
 #[pymethods]
 impl Website {
   /// a new website.
-  pub fn __new__(&self, url: String, raw_content: Option<bool>) -> Self {
+  #[new]
+  pub fn new(url: String, raw_content: Option<bool>) -> Self {
     Website {
       inner: spider::website::Website::new(&url),
       subscription_handles: IndexMap::new(),
       crawl_handles: IndexMap::new(),
       raw_content: raw_content.unwrap_or_default(),
-      collected_data: Box::new(Vec::new()),
       running_in_background: false, // file_handle: None,
     }
   }
@@ -63,49 +43,6 @@ impl Website {
     self.inner.get_status().to_string()
   }
 
-  // /// store data to memory for disk storing. This will create the path if not exist and defaults to ./storage.
-  // pub async fn export_jsonl_data(&self, export_path: Option<String>) -> std::io::Result<()> {
-  //   use spider::tokio::io::AsyncWriteExt;
-  //   let file = match export_path {
-  //     Some(p) => {
-  //       let base_dir = p
-  //         .split("/")
-  //         .into_iter()
-  //         .map(|f| {
-  //           if f.contains(".") {
-  //             "".to_string()
-  //           } else {
-  //             f.to_string()
-  //           }
-  //         })
-  //         .collect::<String>();
-
-  //       spider::tokio::fs::create_dir_all(&base_dir).await?;
-
-  //       if !p.contains(".") {
-  //         p + ".jsonl"
-  //       } else {
-  //         p
-  //       }
-  //     }
-  //     _ => {
-  //       spider::tokio::fs::create_dir_all("./storage").await?;
-  //       "./storage/".to_owned()
-  //         + &self
-  //           .inner
-  //           .get_domain()
-  //           .inner()
-  //           .replace("http://", "")
-  //           .replace("https://", "")
-  //         + "jsonl"
-  //     }
-  //   };
-  //   let mut file = spider::tokio::fs::File::create(file).await?;
-  //   // transform data step needed to auto convert type ..
-  //   file.write_all(&self.collected_data).await?;
-  //   Ok(())
-  // }
-
   /// subscribe and add an event listener.
   pub fn subscribe(mut slf: PyRefMut<'_, Self>, on_page_event: PyObject) -> u32 {
     let mut rx2 = slf
@@ -116,7 +53,7 @@ impl Website {
 
     let handle = pyo3_asyncio::tokio::get_runtime().spawn(async move {
       while let Ok(res) = rx2.recv().await {
-        let page = NPage::new(&res, raw_content);
+        let page = new_page(&res, raw_content);
         Python::with_gil(|py| {
           let _ = on_page_event.call(py, (page, 0), None);
         });
@@ -143,7 +80,7 @@ impl Website {
         match handle {
           Some(h) => {
             h.abort();
-            self.subscription_handles.remove_entry(&id);
+            self.subscription_handles.swap_remove_entry(&id);
             true
           }
           _ => false,
@@ -183,7 +120,7 @@ impl Website {
         match handle {
           Some(h) => {
             h.abort();
-            slf.crawl_handles.remove_entry(&id);
+            slf.crawl_handles.swap_remove_entry(&id);
             true
           }
           _ => false,
@@ -225,9 +162,9 @@ impl Website {
 
           let handle = spider::tokio::spawn(async move {
             while let Ok(res) = rx2.recv().await {
-              let page = NPage::new(&res, raw_content);
+              let page = new_page(&res, raw_content);
               Python::with_gil(|py| {
-                let _ = callback.call(py, (page, 0), None);
+                let _ = callback.call(py, (page,), None);
               });
             }
           });
@@ -258,28 +195,33 @@ impl Website {
             .subscribe(*BUFFER / 2)
             .expect("sync feature should be enabled");
 
-          let handle = pyo3_asyncio::tokio::get_runtime().spawn(async move {
-            while let Ok(res) = rx2.recv().await {
-              Python::with_gil(|py| {
-                let _ = callback.call(py, (NPage::new(&res, raw_content), 0), None);
-              });
-            }
-          });
+          let py: Python<'_> = slf.py();
+          let rt = pyo3_asyncio::tokio::get_runtime();
 
-          let id = match slf.subscription_handles.last() {
-            Some(handle) => handle.0 + 1,
-            _ => 0,
+          let f1 = async {
+            while let Ok(res) = rx2.recv().await {
+              let page = new_page(&res, raw_content);
+              let _ = callback.call(py, (page,), None);
+            }
           };
 
-          slf.subscription_handles.insert(id, handle);
-
-          let _ = pyo3_asyncio::tokio::get_runtime().block_on(async move {
+          let f2 = async {
             if headless {
               slf.inner.crawl().await;
             } else {
               slf.inner.crawl_raw().await;
             }
-            Ok::<(), ()>(())
+          };
+
+          rt.block_on(async move {
+            select! {
+              _ = f1 => {
+                  // println!("Sync receiver droped");
+              }
+              _ = f2 => {
+                  // println!("operation completed");
+              }
+            }
           });
         }
       }
@@ -312,175 +254,266 @@ impl Website {
           });
         }
       }
+    };
+  }
+
+  /// crawl a website smart mode
+  pub fn crawl_smart(
+    mut slf: PyRefMut<'_, Self>,
+    on_page_event: Option<PyObject>,
+    background: Option<bool>,
+  ) {
+    // only run in background if on_page_event is handled for streaming.
+    let background = background.is_some() && background.unwrap_or_default();
+    let raw_content = slf.raw_content;
+
+    if background {
+      slf.running_in_background = background;
+    }
+
+    match on_page_event {
+      Some(callback) => {
+        if background {
+          let mut website = slf.inner.clone();
+          let mut rx2 = website
+            .subscribe(*BUFFER / 2)
+            .expect("sync feature should be enabled");
+
+          let handle = spider::tokio::spawn(async move {
+            while let Ok(res) = rx2.recv().await {
+              let page = new_page(&res, raw_content);
+              Python::with_gil(|py| {
+                let _ = callback.call(py, (page,), None);
+              });
+            }
+          });
+
+          let crawl_id = match slf.crawl_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          let crawl_handle = spider::tokio::spawn(async move {
+            website.crawl_smart().await;
+          });
+
+          let id = match slf.subscription_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          slf.crawl_handles.insert(crawl_id, crawl_handle);
+          slf.subscription_handles.insert(id, handle);
+        } else {
+          let mut rx2 = slf
+            .inner
+            .subscribe(*BUFFER / 2)
+            .expect("sync feature should be enabled");
+
+          let py: Python<'_> = slf.py();
+          let rt = pyo3_asyncio::tokio::get_runtime();
+
+          let f1 = async {
+            while let Ok(res) = rx2.recv().await {
+              let page = new_page(&res, raw_content);
+              let _ = callback.call(py, (page,), None);
+            }
+          };
+
+          let f2 = async {
+            slf.inner.crawl_smart().await;
+          };
+
+          rt.block_on(async move {
+            select! {
+              _ = f1 => {
+                  // println!("Sync receiver droped");
+              }
+              _ = f2 => {
+                  // println!("operation completed");
+              }
+            }
+          });
+        }
+      }
+      _ => {
+        if background {
+          let mut website = slf.inner.clone();
+
+          let crawl_id = match slf.crawl_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          let crawl_handle = spider::tokio::spawn(async move {
+            website.crawl_smart().await;
+          });
+
+          slf.crawl_handles.insert(crawl_id, crawl_handle);
+        } else {
+          let _ = pyo3_asyncio::tokio::get_runtime().block_on(async move {
+            slf.inner.crawl_smart().await;
+            Ok::<(), ()>(())
+          });
+        }
+      }
+    };
+  }
+
+  /// scrape a website
+  pub fn scrape(
+    mut slf: PyRefMut<'_, Self>,
+    on_page_event: Option<PyObject>,
+    background: Option<bool>,
+    headless: Option<bool>,
+  ) {
+    let headless = headless.is_some() && headless.unwrap_or_default();
+    let raw_content = slf.raw_content;
+    let background = background.is_some() && background.unwrap_or_default();
+
+    if background {
+      slf.running_in_background = background;
+    }
+
+    match on_page_event {
+      Some(callback) => {
+        if background {
+          let mut website = slf.inner.clone();
+          let mut rx2 = website
+            .subscribe(*BUFFER / 2)
+            .expect("sync feature should be enabled");
+
+          let handle = spider::tokio::spawn(async move {
+            while let Ok(res) = rx2.recv().await {
+              let page = new_page(&res, raw_content);
+
+              // todo: remove global lock page events
+              Python::with_gil(|py| {
+                let _ = callback.call(py, (page,), None);
+              });
+            }
+          });
+
+          let crawl_id = match slf.crawl_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          let crawl_handle = spider::tokio::spawn(async move {
+            if headless {
+              website.scrape().await;
+            } else {
+              website.scrape_raw().await;
+            }
+          });
+
+          let id = match slf.subscription_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          slf.crawl_handles.insert(crawl_id, crawl_handle);
+          slf.subscription_handles.insert(id, handle);
+        } else {
+          let mut rx2 = slf
+            .inner
+            .subscribe(*BUFFER / 2)
+            .expect("sync feature should be enabled");
+
+          let py: Python<'_> = slf.py();
+          let rt = pyo3_asyncio::tokio::get_runtime();
+
+          let f1 = async {
+            while let Ok(res) = rx2.recv().await {
+              let page = new_page(&res, raw_content);
+              let _ = callback.call(py, (page,), None);
+            }
+          };
+
+          let f2 = async {
+            if headless {
+              slf.inner.scrape().await;
+            } else {
+              slf.inner.scrape_raw().await;
+            }
+          };
+
+          rt.block_on(async move {
+            select! {
+              _ = f1 => {
+                  // println!("Sync receiver droped");
+              }
+              _ = f2 => {
+                  // println!("operation completed");
+              }
+            }
+          });
+        }
+      }
+      _ => {
+        if background {
+          let mut website = slf.inner.clone();
+
+          let crawl_id = match slf.crawl_handles.last() {
+            Some(handle) => handle.0 + 1,
+            _ => 0,
+          };
+
+          let crawl_handle = spider::tokio::spawn(async move {
+            if headless {
+              website.scrape().await;
+            } else {
+              website.scrape_raw().await;
+            }
+          });
+
+          slf.crawl_handles.insert(crawl_id, crawl_handle);
+        } else {
+          let _ = pyo3_asyncio::tokio::get_runtime().block_on(async move {
+            if headless {
+              slf.inner.scrape().await;
+            } else {
+              slf.inner.scrape_raw().await;
+            }
+            Ok::<(), ()>(())
+          });
+        }
+      }
     }
   }
 
-  // /// scrape a website
-  // pub async unsafe fn scrape(
-  //   &mut self,
-  //   // on_page_event: Option<napi::threadsafe_function::ThreadsafeFunction<NPage>>,
-  //   background: Option<bool>,
-  //   headless: Option<bool>,
-  // ) {
-  //   let headless = headless.is_some() && headless.unwrap_or_default();
-  //   let raw_content = self.raw_content;
-  //   let background = background.is_some() && background.unwrap_or_default();
+  /// run a cron job
+  pub fn run_cron(mut slf: PyRefMut<'_, Self>, on_page_event: Option<PyObject>) -> Cron {
+    let cron_handle = match on_page_event {
+      Some(callback) => {
+        let mut rx2 = slf
+          .inner
+          .subscribe(*BUFFER / 2)
+          .expect("sync feature should be enabled");
+        let raw_content = slf.raw_content;
 
-  //   if background {
-  //     self.running_in_background = background;
-  //   }
+        let handler = spider::tokio::spawn(async move {
+          while let Ok(res) = rx2.recv().await {
+            Python::with_gil(|py| {
+              let _ = callback.call(py, (new_page(&res, raw_content),), None);
+            });
+          }
+        });
 
-  //   if background {
-  //     let mut website = self.inner.clone();
+        Some(handler)
+      }
+      _ => None,
+    };
 
-  //     let crawl_id = match self.crawl_handles.last() {
-  //       Some(handle) => handle.0 + 1,
-  //       _ => 0,
-  //     };
+    let inner = pyo3_asyncio::tokio::get_runtime()
+      .block_on(async move {
+        let runner: spider::async_job::Runner = slf.inner.run_cron().await;
+        Ok::<spider::async_job::Runner, ()>(runner)
+      })
+      .unwrap();
 
-  //     let crawl_handle = spider::tokio::spawn(async move {
-  //       if headless {
-  //         website.scrape().await;
-  //       } else {
-  //         website.scrape_raw().await;
-  //       }
-  //     });
-
-  //     self.crawl_handles.insert(crawl_id, crawl_handle);
-  //   } else {
-  //     if headless {
-  //       self.inner.scrape().await;
-  //     } else {
-  //       self.inner.scrape_raw().await;
-  //     }
-  //   }
-
-  //   // match on_page_event {
-  //   //   Some(callback) => {
-  //   //     if background {
-  //   //       let mut website = self.inner.clone();
-  //   //       let mut rx2 = website
-  //   //         .subscribe(*BUFFER / 2)
-  //   //         .expect("sync feature should be enabled");
-
-  //   //       let handle = spider::tokio::spawn(async move {
-  //   //         while let Ok(res) = rx2.recv().await {
-  //   //           callback.call(
-  //   //             Ok(NPage::new(&res, raw_content)),
-  //   //             napi::threadsafe_function::ThreadsafeFunctionCallMode::NonBlocking,
-  //   //           );
-  //   //         }
-  //   //       });
-
-  //   //       let crawl_id = match self.crawl_handles.last() {
-  //   //         Some(handle) => handle.0 + 1,
-  //   //         _ => 0,
-  //   //       };
-
-  //   //       let crawl_handle = spider::tokio::spawn(async move {
-  //   //         if headless {
-  //   //           website.scrape().await;
-  //   //         } else {
-  //   //           website.scrape_raw().await;
-  //   //         }
-  //   //       });
-
-  //   //       let id = match self.subscription_handles.last() {
-  //   //         Some(handle) => handle.0 + 1,
-  //   //         _ => 0,
-  //   //       };
-
-  //   //       self.crawl_handles.insert(crawl_id, crawl_handle);
-  //   //       self.subscription_handles.insert(id, handle);
-  //   //     } else {
-  //   //       let mut rx2 = self
-  //   //         .inner
-  //   //         .subscribe(*BUFFER / 2)
-  //   //         .expect("sync feature should be enabled");
-
-  //   //       let handle = spider::tokio::spawn(async move {
-  //   //         while let Ok(res) = rx2.recv().await {
-  //   //           callback.call(
-  //   //             Ok(NPage::new(&res, raw_content)),
-  //   //             napi::threadsafe_function::ThreadsafeFunctionCallMode::NonBlocking,
-  //   //           );
-  //   //         }
-  //   //       });
-
-  //   //       if headless {
-  //   //         self.inner.scrape().await;
-  //   //       } else {
-  //   //         self.inner.scrape_raw().await;
-  //   //       }
-
-  //   //       let id = match self.subscription_handles.last() {
-  //   //         Some(handle) => handle.0 + 1,
-  //   //         _ => 0,
-  //   //       };
-
-  //   //       self.subscription_handles.insert(id, handle);
-  //   //     }
-  //   //   }
-  //   //   _ => {
-  //   //     if background {
-  //   //       let mut website = self.inner.clone();
-
-  //   //       let crawl_id = match self.crawl_handles.last() {
-  //   //         Some(handle) => handle.0 + 1,
-  //   //         _ => 0,
-  //   //       };
-
-  //   //       let crawl_handle = spider::tokio::spawn(async move {
-  //   //         if headless {
-  //   //           website.scrape().await;
-  //   //         } else {
-  //   //           website.scrape_raw().await;
-  //   //         }
-  //   //       });
-
-  //   //       self.crawl_handles.insert(crawl_id, crawl_handle);
-  //   //     } else {
-  //   //       if headless {
-  //   //         self.inner.scrape().await;
-  //   //       } else {
-  //   //         self.inner.scrape_raw().await;
-  //   //       }
-  //   //     }
-  //   //   }
-  //   // }
-  // }
-
-  //  /// run a cron job
-  // pub async unsafe fn run_cron(
-  //   &mut self,
-  //   on_page_event: Option<napi::threadsafe_function::ThreadsafeFunction<NPage>>,
-  // ) -> Cron {
-  //   let cron_handle = match on_page_event {
-  //     Some(callback) => {
-  //       let mut rx2 = self
-  //         .inner
-  //         .subscribe(*BUFFER / 2)
-  //         .expect("sync feature should be enabled");
-  //       let raw_content = self.raw_content;
-
-  //       let handler = spider::tokio::spawn(async move {
-  //         while let Ok(res) = rx2.recv().await {
-  //           callback.call(
-  //             Ok(NPage::new(&res, raw_content)),
-  //             napi::threadsafe_function::ThreadsafeFunctionCallMode::NonBlocking,
-  //           );
-  //         }
-  //       });
-
-  //       Some(handler)
-  //     }
-  //     _ => None,
-  //   };
-
-  //   let inner = self.inner.run_cron().await;
-
-  //   Cron { inner, cron_handle }
-  // }
+    Cron { inner, cron_handle }
+  }
 
   /// get all the links of a website
   pub fn get_links(&self) -> Vec<String> {
@@ -494,8 +527,34 @@ impl Website {
   }
 
   /// get the size of the website in amount of pages crawled. If you ran the page in the background, this value will not update.
-  pub fn size(&mut self) -> u32 {
-    self.inner.size() as u32
+  #[getter]
+  pub fn size(slf: PyRef<'_, Self>) -> u32 {
+    slf.inner.size() as u32
+  }
+
+  /// get the configuration custom HTTP headers
+  pub fn get_configuration_headers(&self) -> Vec<(String, String)> {
+    let mut map = Vec::new();
+
+    match self.inner.configuration.headers.as_ref() {
+      Some(h) => {
+        for v in h.iter() {
+          let mut value = String::new();
+
+          match v.1.to_str() {
+            Ok(vv) => {
+              value.push_str(vv);
+            }
+            _ => (),
+          };
+
+          map.push((v.0.to_string(), value))
+        }
+      }
+      _ => (),
+    }
+
+    map
   }
 
   /// get all the pages of a website - requires calling website.scrape
@@ -506,7 +565,7 @@ impl Website {
     match self.inner.get_pages() {
       Some(p) => {
         for page in p.iter() {
-          pages.push(NPage::new(page, raw_content));
+          pages.push(new_page(page, raw_content));
         }
       }
       _ => (),
@@ -531,44 +590,49 @@ impl Website {
     self.inner.clear();
   }
 
-  // /// Set HTTP headers for request using [reqwest::header::HeaderMap](https://docs.rs/reqwest/latest/reqwest/header/struct.HeaderMap.html).
-  // pub fn with_headers(mut slf: PyRefMut<'_, Self>, headers: Option<Object>) -> PyRefMut<'_, Self> {
-  //   use std::str::FromStr;
+  /// Set HTTP headers for request using [reqwest::header::HeaderMap](https://docs.rs/reqwest/latest/reqwest/header/struct.HeaderMap.html).
+  pub fn with_headers(
+    mut slf: PyRefMut<'_, Self>,
+    headers: Option<PyObject>,
+  ) -> PyRefMut<'_, Self> {
+    use std::str::FromStr;
+    match headers {
+      Some(obj) => {
+        let mut h = spider::reqwest::header::HeaderMap::new();
+        let py = slf.py();
+        let dict = obj.downcast::<pyo3::types::PyDict>(py);
 
-  //   match headers {
-  //     Some(obj) => {
-  //       let mut h = spider::reqwest::header::HeaderMap::new();
-  //       let keys = Object::keys(&obj).unwrap_or_default();
+        match dict {
+          Ok(keys) => {
+            for key in keys.into_iter() {
+              let header_key = spider::reqwest::header::HeaderName::from_str(&key.0.to_string());
 
-  //       for key in keys.into_iter() {
-  //         let header_key = spider::reqwest::header::HeaderName::from_str(&key);
+              match header_key {
+                Ok(hn) => {
+                  let header_value = key.1.to_string();
 
-  //         match header_key {
-  //           Ok(hn) => {
-  //             let header_value = obj
-  //               .get::<String, String>(key)
-  //               .unwrap_or_default()
-  //               .unwrap_or_default();
+                  match spider::reqwest::header::HeaderValue::from_str(&header_value) {
+                    Ok(hk) => {
+                      h.append(hn, hk);
+                    }
+                    _ => (),
+                  }
+                }
+                _ => (),
+              }
+            }
+            slf.inner.with_headers(Some(h));
+          }
+          _ => (),
+        }
+      }
+      _ => {
+        slf.inner.with_headers(None);
+      }
+    };
 
-  //             match spider::reqwest::header::HeaderValue::from_str(&header_value) {
-  //               Ok(hk) => {
-  //                 h.append(hn, hk);
-  //               }
-  //               _ => (),
-  //             }
-  //           }
-  //           _ => (),
-  //         }
-  //       }
-  //       slf.inner.with_headers(Some(h));
-  //     }
-  //     _ => {
-  //       slf.inner.with_headers(None);
-  //     }
-  //   };
-
-  //   slf
-  // }
+    slf
+  }
 
   /// Add user agent to request.
   pub fn with_user_agent(
@@ -633,6 +697,18 @@ impl Website {
     slf
   }
 
+  /// Set a crawl depth limit. If the value is 0 there is no limit.
+  pub fn with_depth(mut slf: PyRefMut<'_, Self>, depth: usize) -> PyRefMut<'_, Self> {
+    slf.inner.with_depth(depth);
+    slf
+  }
+
+  /// Cache the page following HTTP rules.
+  pub fn with_caching(mut slf: PyRefMut<'_, Self>, cache: bool) -> PyRefMut<'_, Self> {
+    slf.inner.with_caching(cache);
+    slf
+  }
+
   /// add external domains
   pub fn with_external_domains(
     mut slf: PyRefMut<'_, Self>,
@@ -654,18 +730,11 @@ impl Website {
 
     match budget {
       Some(d) => {
-        let v = d
-          .iter()
-          .map(|e| e.0.to_owned() + "," + &e.1.to_string())
-          .collect::<String>();
-        let v = v
-          .split(",")
-          .collect::<Vec<_>>()
-          .chunks(2)
-          .map(|x| (x[0], x[1].parse::<u32>().unwrap_or_default()))
-          .collect::<HashMap<&str, u32>>();
-
-        slf.inner.with_budget(Some(v));
+        slf.inner.with_budget(Some(
+          d.iter()
+            .map(|(k, v)| (k.as_str(), *v))
+            .collect::<HashMap<&str, u32>>(),
+        ));
       }
       _ => (),
     }
@@ -673,25 +742,101 @@ impl Website {
     slf
   }
 
+  /// Take a screenshot of the page when using chrome.
+  pub fn with_screenshot<'a>(
+    mut slf: PyRefMut<'a, Self>,
+    screenshot_configs: Option<&'a PyDict>,
+  ) -> PyRefMut<'a, Self> {
+    if let Some(py_obj) = screenshot_configs {
+      if let Ok(config_json) = pydict_to_json_value(py_obj) {
+        match serde_json::from_value::<spider::configuration::ScreenShotConfig>(config_json) {
+          Ok(configs) => {
+            slf.inner.with_screenshot(Some(configs));
+          }
+          Err(e) => {
+            spider::utils::log("", e.to_string());
+          }
+        }
+      } else {
+        spider::utils::log("Error extracting String from PyAny", "");
+      }
+    }
+
+    slf
+  }
+
+  /// Use OpenAI to generate dynamic javascript snippets. Make sure to set the `OPENAI_API_KEY` env variable.
+  pub fn with_openai<'a>(
+    mut slf: PyRefMut<'a, Self>,
+    openai_configs: Option<&'a PyDict>,
+  ) -> PyRefMut<'a, Self> {
+    if let Some(py_obj) = openai_configs {
+      if let Ok(config_json) = pydict_to_json_value(py_obj) {
+        match serde_json::from_value::<spider::configuration::GPTConfigs>(config_json) {
+          Ok(configs) => {
+            slf.inner.with_openai(Some(configs));
+          }
+          Err(e) => {
+            spider::utils::log("", e.to_string());
+          }
+        }
+      } else {
+        spider::utils::log("Error extracting String from PyAny for OpenAI config", "");
+      }
+    }
+
+    slf
+  }
+
   /// Regex black list urls from the crawl
   pub fn with_blacklist_url(
     mut slf: PyRefMut<'_, Self>,
     blacklist_url: Option<Vec<String>>,
+  ) -> PyRefMut<'_, Self> {
+    slf.inner.with_blacklist_url(match blacklist_url {
+      Some(v) => {
+        let mut blacklist: Vec<CompactString> = Vec::new();
+        for item in v {
+          blacklist.push(CompactString::new(item));
+        }
+        Some(blacklist)
+      }
+      _ => None,
+    });
+
+    slf
+  }
+
+  /// Use network interception for the request to only allow content that matches the host. If the content is from a 3rd party it needs to be part of our include list.
+  pub fn with_chrome_intercept(
+    mut slf: PyRefMut<'_, Self>,
+    chrome_intercept: bool,
+    block_images: bool,
   ) -> PyRefMut<'_, Self> {
     slf
       .inner
-      .configuration
-      .with_blacklist_url(match blacklist_url {
-        Some(v) => {
-          let mut blacklist: Vec<CompactString> = Vec::new();
-          for item in v {
-            blacklist.push(CompactString::new(item));
-          }
-          Some(blacklist)
-        }
-        _ => None,
-      });
+      .with_chrome_intercept(chrome_intercept, block_images);
+    slf
+  }
 
+  /// Wait for idle network request. This method does nothing if the `chrome` feature is not enabled.
+  /// Set the timeout to 0 to disable the timeout.
+  pub fn with_wait_for_idle_network(
+    mut slf: PyRefMut<'_, Self>,
+    wait_for_idle_network: bool,
+    timeout: u64,
+  ) -> PyRefMut<'_, Self> {
+    slf
+      .inner
+      .with_wait_for_idle_network(if wait_for_idle_network {
+        Some(WaitForIdleNetwork::new(if timeout == 0 {
+          None
+        } else {
+          Some(core::time::Duration::from_millis(timeout))
+        }))
+      } else {
+        None
+      });
     slf
   }
 
@@ -727,6 +872,12 @@ impl Website {
     slf
   }
 
+  /// Use stealth mode for the request. This does nothing without the [chrome] flag enabled.
+  pub fn with_stealth(mut slf: PyRefMut<'_, Self>, stealth_mode: bool) -> PyRefMut<'_, Self> {
+    slf.inner.configuration.with_stealth(stealth_mode);
+    slf
+  }
+
   /// build the inner website - not required for all builder_steps
   pub fn build(mut slf: PyRefMut<'_, Self>) -> PyRefMut<'_, Self> {
     match slf.inner.build() {
@@ -738,6 +889,7 @@ impl Website {
 }
 
 /// a runner for handling crons
+#[pyclass]
 pub struct Cron {
   /// the runner task
   inner: spider::async_job::Runner,
@@ -745,14 +897,17 @@ pub struct Cron {
   cron_handle: Option<JoinHandle<()>>,
 }
 
+#[pymethods]
 impl Cron {
   /// stop the cron instance
-
-  pub async unsafe fn stop(&mut self) {
-    self.inner.stop().await;
-    match &self.cron_handle {
+  pub fn stop(mut slf: PyRefMut<'_, Self>) {
+    match &slf.cron_handle {
       Some(h) => h.abort(),
       _ => (),
-    }
+    };
+    let _ = pyo3_asyncio::tokio::get_runtime().block_on(async move {
+      slf.inner.stop().await;
+      Ok::<(), ()>(())
+    });
   }
 }
